from flask import Flask

from flask_discord_interactions import DiscordInteractions
from flask_discord_interactions.context import ApplicationCommandType

def test_register_command():
    app = Flask(__name__)
    app.config["DONT_VALIDATE_SIGNATURE"] = True
    app.config["DONT_REGISTER_WITH_DISCORD"] = True

    discord = DiscordInteractions(app)

    @discord.command()
    def ping(ctx):
        return "pong"

    discord.update_commands()


def test_register_user_command():
    app = Flask(__name__)
    app.config["DONT_VALIDATE_SIGNATURE"] = True
    app.config["DONT_REGISTER_WITH_DISCORD"] = True

    discord = DiscordInteractions(app)

    @discord.command(type=ApplicationCommandType.USER)
    def ping(ctx):
        return "pong"

    @discord.command(type=ApplicationCommandType.USER)
    def PING(ctx):
        return "pong"

    @discord.command(name="user test", type=ApplicationCommandType.USER)
    def ping(ctx):
        return "pong"

    discord.update_commands()


def test_register_message_command():
    app = Flask(__name__)
    app.config["DONT_VALIDATE_SIGNATURE"] = True
    app.config["DONT_REGISTER_WITH_DISCORD"] = True

    discord = DiscordInteractions(app)

    @discord.command(type=ApplicationCommandType.MESSAGE)
    def ping(ctx):
        return "pong"

    @discord.command(type=ApplicationCommandType.MESSAGE)
    def PING(ctx):
        return "pong"

    @discord.command(name="user test", type=ApplicationCommandType.MESSAGE)
    def ping(ctx):
        return "pong"

    discord.update_commands()


def test_register_subcommand():
    app = Flask(__name__)
    app.config["DONT_VALIDATE_SIGNATURE"] = True
    app.config["DONT_REGISTER_WITH_DISCORD"] = True

    discord = DiscordInteractions(app)

    group = discord.command_group("group")

    @group.command()
    def subcommand(ctx):
        return "pong"

    discord.update_commands()


def test_register_options():
    app = Flask(__name__)
    app.config["DONT_VALIDATE_SIGNATURE"] = True
    app.config["DONT_REGISTER_WITH_DISCORD"] = True

    discord = DiscordInteractions(app)

    @discord.command()
    def ping(ctx, option1: str, option2: float, option3: str = ""):
        return f"pong"

    discord.update_commands()


def test_register_subcommand_options():
    app = Flask(__name__)
    app.config["DONT_VALIDATE_SIGNATURE"] = True
    app.config["DONT_REGISTER_WITH_DISCORD"] = True

    discord = DiscordInteractions(app)

    group = discord.command_group("group")

    @group.command()
    def subcommand(ctx, option1: str, option2: float, option3: str = ""):
        return "pong"

    discord.update_commands()



import dataclasses
from flask_discord_interactions.models.user import Member
import inspect
from typing import List, Union
import json
from datetime import datetime

from flask_discord_interactions.models.utils import LoadableDataclass
from flask_discord_interactions.models.component import Component
from flask_discord_interactions.models.embed import Embed


class ResponseType:
    "Represents the different response type integers."
    PONG = 1
    CHANNEL_MESSAGE_WITH_SOURCE = 4
    DEFERRED_CHANNEL_MESSAGE_WITH_SOURCE = 5
    DEFERRED_UPDATE_MESSAGE = 6
    UPDATE_MESSAGE = 7
    APPLICATION_COMMAND_AUTOCOMPLETE_RESULT = 8


@dataclasses.dataclass
class Message(LoadableDataclass):
    """
    Represents a Message, often a response to a Discord interaction (incoming
    webhook) or a followup message (outgoing webhook).

    Attributes
    ----------

    content
        The message body/content.
    tts
        Whether the message should be sent with text-to-speech.
    embed
        The embed object (dict) accompanying the message.
    embeds
        An array of embed objects. Specify just one of ``embed`` or ``embeds``.
    allowed_mentions
        An object representing which roles and users the message is allowed
        to mention. See the Discord API docs for details.
    deferred
        Whether the message should be deferred (display a loading state
        to the user temporarily). Only valid for incoming webhooks.
    ephemeral
        Whether the message should be ephemeral (only displayed temporarily
        to only the user who used the command). Only valid for incoming
        webhooks.
    update
        Whether to update the initial message. Only valid for Message Component
        interactions / custom ID handlers.
    file
        The file to attach to the message. Only valid for outgoing webhooks.
        Pass a 2-tuple containing the file name and a file object, e.g.

        .. code-block:: python

            ("README.md", open("README.md", "rb"))

        See the ``files`` parameter of :func:`requests.request` for details.
    files
        A list of files to attach to the message. Specify just one of
        ``file`` or ``files``. Only valid for outgoing webhooks.
    components
        An array of :class:`.Component` objects representing message
        components.
    """

    content: str = None
    tts: bool = False
    embed: Union[Embed, dict] = None
    embeds: List[Union[Embed, dict]] = None
    allowed_mentions: dict = dataclasses.field(
        default_factory=lambda: {"parse": ["roles", "users", "everyone"]}
    )
    deferred: bool = False
    ephemeral: bool = False
    update: bool = False
    file: tuple = None
    files: List[tuple] = None
    components: List[Component] = None

    # These fields are only set on incoming messages
    id: str = None
    channel_id: str = None
    timestamp: datetime = None
    edited_timestamp: datetime = None
    author: Member = None

    def __post_init__(self):
        if self.embed is not None and self.embeds is not None:
            raise ValueError("Specify only one of embed or embeds")
        if self.embed is not None:
            self.embeds = [self.embed]

        if self.file is not None and self.files is not None:
            raise ValueError("Specify only one of file or files")
        if self.file is not None:
            self.files = [self.file]

        if self.ephemeral and self.files is not None:
            raise ValueError("Ephemeral Messages cannot include files.")

        if self.embeds is not None:
            for i, embed in enumerate(self.embeds):
                if not dataclasses.is_dataclass(embed):
                    self.embeds[i] = Embed.from_dict(embed)

        if self.update:
            if self.deferred:
                self.Message_type = ResponseType.DEFERRED_UPDATE_MESSAGE
            else:
                self.Message_type = ResponseType.UPDATE_MESSAGE
        else:
            if self.deferred:
                self.Message_type = ResponseType.DEFERRED_CHANNEL_MESSAGE_WITH_SOURCE
            else:
                self.Message_type = ResponseType.CHANNEL_MESSAGE_WITH_SOURCE

        if isinstance(self.timestamp, str):
            self.timestamp = datetime.strptime(self.timestamp, "%Y-%m-%dT%H:%M:%S.%f%z")

        if isinstance(self.edited_timestamp, str):
            self.edited_timestamp = datetime.strptime(
                self.edited_timestamp, "%Y-%m-%dT%H:%M:%S.%f%z"
            )

        if isinstance(self.author, dict):
            self.author = Member.from_dict(self.author)

    @property
    def flags(self):
        """
        The flags sent with this Message, determined by whether it is
        ephemeral.
        """
        return 64 if self.ephemeral else 0

    def dump_embeds(self):
        "Returns the embeds of this Message as a list of dicts."
        return (
            [embed.dump() for embed in self.embeds] if self.embeds is not None else None
        )

    def dump_components(self):
        "Returns the message components as a list of dicts."
        return (
            [c.dump() for c in self.components] if self.components is not None else None
        )

    @classmethod
    def from_return_value(cls, result):
        """
        Convert a function return value into a Message object.
        Converts ``None`` to an empty Message, or any other type to ``str``
        as message content.

        Parameters
        ----------
        result
            The function return value to convert into a ``Message`` object.
        """

        async def construct_async(result):
            return cls.from_return_value(await result)

        if result is None:
            return cls()
        elif inspect.isawaitable(result):
            return construct_async(result)
        elif isinstance(result, cls):
            return result
        else:
            return cls(str(result))

    def dump(self):
        """
        Return this ``Message`` as a dict to be sent in Message to an
        incoming webhook.
        """

        if (
            self.content is None
            and self.embeds is None
            and self.files is None
            and not self.deferred
        ):
            raise ValueError(
                "Supply at least one of content, embeds, files, or deferred."
            )

        if self.files:
            raise ValueError("files are not allowed in an initial Interaction Message")

        if self.update:
            raise ValueError("update is only valid for custom ID handlers")

        return {
            "type": self.Message_type,
            "data": {
                "content": self.content,
                "tts": self.tts,
                "embeds": self.dump_embeds(),
                "allowed_mentions": self.allowed_mentions,
                "flags": self.flags,
                "components": self.dump_components(),
            },
        }

    def dump_handler(self):
        """
        Return this ``Message`` as a dict to be sent in reply to a Message
        Component interaction.
        """

        if self.files:
            raise ValueError("files are not allowed in a custom handler Message")

        return {
            "type": self.Message_type,
            "data": {
                "content": self.content,
                "tts": self.tts,
                "embeds": self.dump_embeds(),
                "allowed_mentions": self.allowed_mentions,
                "flags": self.flags,
                "components": self.dump_components(),
            },
        }

    def dump_followup(self):
        """
        Return this ``Message`` as a dict to be sent to an outgoing webhook.
        """

        if (
            self.content is None
            and self.embeds is None
            and self.files is None
            and not self.deferred
        ):
            raise ValueError(
                "Supply at least one of content, embeds, files, or deferred."
            )

        if self.ephemeral or self.deferred or self.update:
            raise ValueError(
                "ephemeral and deferred are not valid in followup Messages"
            )

        return {
            "content": self.content,
            "tts": self.tts,
            "embeds": self.dump_embeds(),
            "allowed_mentions": self.allowed_mentions,
            "components": self.dump_components(),
        }

    def dump_multipart(self):
        """
        Return this ``Message`` to be sent to an outgoing webhook.
        Handles multipart encoding for file attachments.

        Returns an object that may have ``data``, ``files``, or ``json``
        fields.
        """

        if self.files:
            payload_json = json.dumps(self.dump_followup())

            multipart = []
            for file in self.files:
                multipart.append(("file", file))

            return {"data": {"payload_json": payload_json}, "files": multipart}
        else:
            return {"json": self.dump_followup()}



import threading

from flask import Flask

from flask_discord_interactions import DiscordInteractions, ResponseType, InteractionType, Message


def test_flask():
    app = Flask(__name__)
    app.config["DONT_VALIDATE_SIGNATURE"] = True
    app.config["DONT_REGISTER_WITH_DISCORD"] = True

    discord = DiscordInteractions(app)

    @discord.command()
    def ping(ctx, pong: str = "ping"):
        return f"Ping {pong}!"

    discord.set_route("/interactions")

    with app.test_client() as client:
        response = client.post("/interactions", json={
            "type": InteractionType.APPLICATION_COMMAND,
            "id": 1,
            "channel_id": "",
            "guild_id": "",
            "token": "",
            "data": {
                "id": 1,
                "name": "ping",
                "options": [
                    {
                        "type": 1,
                        "name": "Pong"
                    }
                ]
            },
            "member": {
                "id": 1,
                "nick": "",
                "user": {
                    "id": 1,
                    "username": "test"
                }
            }
        })

        assert response.status_code == 200
        assert response.get_json()["type"] == \
            ResponseType.CHANNEL_MESSAGE_WITH_SOURCE

        assert response.get_json()["data"]["content"] == "Ping Pong!"

def test_app_factory():
    discord = DiscordInteractions()

    @discord.command()
    def ping(ctx, pong: str = "ping"):
        return f"Ping {pong}!"

    def create_app():
        app = Flask(__name__)
        discord.init_app(app)
        discord.set_route("/interactions", app=app)
        return app

    app = create_app()
    app.config["DONT_VALIDATE_SIGNATURE"] = True
    app.config["DONT_REGISTER_WITH_DISCORD"] = True

    with app.test_client() as client:
        response = client.post("/interactions", json={
            "type": InteractionType.APPLICATION_COMMAND,
            "id": 1,
            "channel_id": "",
            "guild_id": "",
            "token": "",
            "data": {
                "id": 1,
                "name": "ping",
                "options": [
                    {
                        "type": 1,
                        "name": "Pong"
                    }
                ]
            },
            "member": {
                "id": 1,
                "nick": "",
                "user": {
                    "id": 1,
                    "username": "test"
                }
            }
        })

        assert response.status_code == 200
        assert response.get_json()["type"] == \
            ResponseType.CHANNEL_MESSAGE_WITH_SOURCE

        assert response.get_json()["data"]["content"] == "Ping Pong!"

def test_followup():
    app = Flask(__name__)
    app.config["DONT_VALIDATE_SIGNATURE"] = True
    app.config["DONT_REGISTER_WITH_DISCORD"] = True

    discord = DiscordInteractions(app)

    ref_to_thread = None

    @discord.command()
    def ping(ctx):
        nonlocal ref_to_thread

        def do_followup():
            ctx.edit("hi")

        ref_to_thread = thread = threading.Thread(target=do_followup)
        thread.start()

        return Message(deferred=True)


    discord.set_route("/interactions")

    with app.test_client() as client:
        response = client.post("/interactions", json={
            "type": InteractionType.APPLICATION_COMMAND,
            "id": 1,
            "channel_id": "",
            "guild_id": "",
            "token": "",
            "data": {
                "id": 1,
                "name": "ping",
            }
        })

    # Make sure we wait for the thread to complete
    ref_to_thread.join()



from flask_discord_interactions import (Message, ResponseType, ActionRow,
                                        Button, ComponentType, ButtonStyles,
                                        SelectMenu, SelectMenuOption)


def test_parse_arguments(discord, client):
    @discord.custom_handler()
    def handler(ctx, string_arg, int_arg: int):
        return f"String: {string_arg}, type(int_arg): {type(int_arg)}"

    assert client.run_handler(handler, "hello!", "42").content == \
        "String: hello!, type(int_arg): <class 'int'>"


def test_action_row(discord, client):
    @discord.command()
    def action_row(ctx):
        return Message(
            content="Hi!",
            components=[
                ActionRow()
            ]
        )

    expected = {
        "type": ResponseType.CHANNEL_MESSAGE_WITH_SOURCE,
        "data": {
            "content": "Hi!",
            "embeds": None,
            "flags": 0,
            "tts": False,
            "allowed_mentions": {"parse": ["roles", "users", "everyone"]},
            "components": [{
                "type": ComponentType.ACTION_ROW
            }]
        }
    }

    assert client.run("action_row").dump() == expected


def test_button(discord, client):
    @discord.command()
    def button(ctx):
        return Message(
            content="Hi!",
            components=[
                ActionRow(components=[
                    Button(
                        style=ButtonStyles.PRIMARY,
                        custom_id="my_button",
                        label="My Button"
                    )
                ])
            ]
        )

    expected = {
        "type": ResponseType.CHANNEL_MESSAGE_WITH_SOURCE,
        "data": {
            "content": "Hi!",
            "embeds": None,
            "flags": 0,
            "tts": False,
            "allowed_mentions": {"parse": ["roles", "users", "everyone"]},
            "components": [{
                "type": ComponentType.ACTION_ROW,
                "components": [{
                    "type": ComponentType.BUTTON,
                    "style": ButtonStyles.PRIMARY,
                    "custom_id": "my_button",
                    "label": "My Button",
                    "disabled": False
                }]
            }]
        }
    }

    assert client.run("button").dump() == expected


def test_select_menu(discord, client):
    @discord.command()
    def selectmenu(ctx):
        return Message(
            content="Hi!",
            components=[
                ActionRow(components=[
                    SelectMenu(
                        custom_id="my_menu",
                        placeholder="Choose an option",
                        options=[
                            SelectMenuOption(
                                label="Option 1",
                                value="option1"
                            ),
                            SelectMenuOption(
                                label="Option 2",
                                value="option2"
                            )
                        ]
                    )
                ])
            ]
        )

    expected = {
        "type": ResponseType.CHANNEL_MESSAGE_WITH_SOURCE,
        "data": {
            "content": "Hi!",
            "embeds": None,
            "flags": 0,
            "tts": False,
            "allowed_mentions": {"parse": ["roles", "users", "everyone"]},
            "components": [{
                "type": ComponentType.ACTION_ROW,
                "components": [{
                    "type": ComponentType.SELECT_MENU,
                    "custom_id": "my_menu",
                    "placeholder": "Choose an option",
                    "disabled": False,
                    "options": [
                        {
                            "label": "Option 1",
                            "value": "option1",
                            "default": False,
                        },
                        {
                            "label": "Option 2",
                            "value": "option2",
                            "default": False,
                        }
                    ],
                    "max_values": 1,
                    "min_values": 1,
                }]
            }]
        }
    }

    assert client.run("selectmenu").dump() == expected



from flask_discord_interactions import CommandOptionType, Context, Member


def test_subcommand(discord, client):
    group = discord.command_group("group")

    @group.command()
    def sub_one(ctx):
        return "sub one"

    @group.command()
    def sub_two(ctx):
        return "sub two"

    assert client.run("group", "sub_one").content == "sub one"
    assert client.run("group", "sub_two").content == "sub two"


def test_subcommand_groups(discord, client):
    group = discord.command_group("group")

    subgroup_one = group.subgroup("sub_one")
    subgroup_two = group.subgroup("sub_two")

    @subgroup_one.command()
    def sub_one_one(ctx):
        return "one one"

    @subgroup_one.command()
    def sub_one_two(ctx):
        return "one two"

    @subgroup_two.command()
    def sub_two_one(ctx):
        return "two one"

    assert client.run("group", "sub_one", "sub_one_one").content == "one one"
    assert client.run("group", "sub_one", "sub_one_two").content == "one two"
    assert client.run("group", "sub_two", "sub_two_one").content == "two one"


def test_oldstyle_subcommand(discord, client):
    @discord.command(options=[
        {
            "name": "google",
            "description": "Search with Google",
            "type": CommandOptionType.SUB_COMMAND,
            "options": [{
                "name": "query",
                "description": "Search query",
                "type": CommandOptionType.STRING,
                "required": True
            }]
        },
        {
            "name": "bing",
            "description": "Search with Bing",
            "type": CommandOptionType.SUB_COMMAND,
            "options": [{
                "name": "query",
                "description": "Search query",
                "type": CommandOptionType.STRING,
                "required": True
            }]
        }
    ])
    def search(ctx, subcommand, *, query):
        "Search the Internet!"
        if subcommand == "google":
            return f"https://google.com/search?q={query}"
        if subcommand == "bing":
            return f"https://bing.com/search?q={query}"

    assert client.run("search", "google", query="hello").content == \
        "https://google.com/search?q=hello"
    assert client.run("search", "bing", query="hello").content == \
        "https://bing.com/search?q=hello"

def test_context_with_subcommand(discord, client):
    group = discord.command_group("group")

    @group.command()
    def subcommand(ctx):
        return ctx.author.display_name

    context = Context(author=Member(username="Bob"))

    with client.context(context):
        assert client.run("group", "subcommand").content == "Bob"



import re
import enum
import inspect
import itertools

from flask_discord_interactions.context import Context, AsyncContext
from flask_discord_interactions.models import (
    Message,
    CommandOptionType,
    ApplicationCommandType,
    User,
    Member,
    Channel,
    Role,
    Autocomplete,
    Option,
)


class Command:
    """
    Represents an Application Command.

    Attributes
    ----------
    command
        Function to call when the command is invoked.
    name
        Name for this command (appears in the Discord client). If omitted,
        infers the name based on the name of the function.
    description
        Description for this command (appears in the Discord client). If
        omitted, infers the description based on the docstring of the function,
        or sets the description to "No description", if ``ApplicationCommandType``
        is ``CHAT_INPUT``, else set description to ``None``.
    options
        Array of options that can be passed to this command. If omitted,
        infers the options based on the function parameters and type
        annotations.
    annotations
        Dictionary of descriptions for each option provided. Use this only if
        you want the options to be inferred from the parameters and type
        annotations. Do not use with ``options``. If omitted, and if
        ``options`` is not provided, option descriptions default to
        "No description".
    command_type
        Type for this command (depend on the action in the Discord client).
        The value is in ``ApplicationCommandType``. If omitted, set the default
        value to ``ApplicationCommandType.CHAT_INPUT``.
    default_permission
        Whether the command is enabled by default. Default is True.
    permissions
        List of permission overwrites.
    discord
        DiscordInteractionsBlueprint instance which this Command is associated
        with.
    """

    def __init__(
        self,
        command,
        name,
        description,
        options,
        annotations,
        command_type=ApplicationCommandType.CHAT_INPUT,
        default_permission=True,
        permissions=None,
        discord=None,
    ):
        self.command = command
        self.name = name
        self.description = description
        self.options = options
        self.annotations = annotations or {}
        self.type = command_type
        self.default_permission = default_permission
        self.permissions = permissions or []
        self.discord = discord

        if self.name is None:
            self.name = command.__name__

        if not 1 <= len(self.name) <= 32:
            raise ValueError(
                f"Error adding command {self.name}: "
                "Command name must be between 1 and 32 characters."
            )
        if self.type is ApplicationCommandType.CHAT_INPUT:
            if self.description is None:
                self.description = command.__doc__ or "No description"
            if self.name != self.name.lower():
                raise ValueError(
                    f"Error adding command {self.name}: "
                    "Command name must be fully lowercase. "
                    "No UPPERCASE or CamelCase names are allowed."
                )
            if not re.fullmatch(r"^[\w-]{1,32}$", self.name):
                raise ValueError(
                    f"Error adding command {self.name}: "
                    "Command name does not match regex. "
                    "(Perhaps it contains an invalid character?)"
                )
            if not 1 <= len(self.description) <= 100:
                raise ValueError(
                    f"Error adding command {self.name}: "
                    "Command description must be between 1 and 100 characters."
                )
        else:
            self.description = None

        self.is_async = inspect.iscoroutinefunction(self.command)

        if self.options:
            self.options = [
                (o.dump() if isinstance(o, Option) else o) for o in self.options
            ]

        if self.type is ApplicationCommandType.CHAT_INPUT and self.options is None:
            sig = inspect.signature(self.command)

            self.options = []
            for parameter in itertools.islice(sig.parameters.values(), 1, None):

                annotation = parameter.annotation
                autocomplete = False

                if type(annotation) == Autocomplete:
                    annotation = annotation.t
                    autocomplete = True

                # Primitive Types
                if annotation == int:
                    ptype = CommandOptionType.INTEGER
                elif annotation == bool:
                    ptype = CommandOptionType.BOOLEAN
                elif annotation == str:
                    ptype = CommandOptionType.STRING
                elif annotation == float:
                    ptype = CommandOptionType.NUMBER

                # Discord Models
                elif annotation in [User, Member]:
                    ptype = CommandOptionType.USER
                elif annotation == Channel:
                    ptype = CommandOptionType.CHANNEL
                elif annotation == Role:
                    ptype = CommandOptionType.ROLE

                # Enums (Used with choices)
                elif issubclass(annotation, enum.IntEnum):
                    ptype = CommandOptionType.INTEGER
                elif issubclass(annotation, enum.Enum):
                    ptype = CommandOptionType.STRING

                else:
                    raise ValueError(f"Invalid type annotation {annotation}")

                option = {
                    "name": parameter.name,
                    "description": self.annotations.get(
                        parameter.name, "No description"
                    ),
                    "type": ptype,
                    "required": (parameter.default == parameter.empty),
                    "autocomplete": autocomplete,
                }

                if issubclass(annotation, enum.Enum):
                    choices = []

                    if issubclass(annotation, enum.IntEnum):
                        value_type = int
                    else:
                        value_type = str

                    for name, member in annotation.__members__.items():
                        choices.append(
                            {"name": name, "value": value_type(member.value)}
                        )

                    option["choices"] = choices

                self.options.append(option)

    def make_context_and_run(self, discord, app, data):
        """
        Creates the :class:`Context` object for an invocation of this
        command, then invokes itself.

        Parameters
        ----------
        discord
            The :class:`DiscordInteractions` object used to receive this
            interaction.
        app
            The Flask app used to receive this interaction.
        data
            The incoming interaction data.

        Returns
        -------
        Message
            The response by the command, converted to a Message object.
        """

        if self.is_async:
            context = AsyncContext.from_data(discord, app, data)
        else:
            context = Context.from_data(discord, app, data)
        args, kwargs = context.create_args()

        result = self.run(context, *args, **kwargs)

        return Message.from_return_value(result)

    def run(self, context, *args, **kwargs):
        """
        Invokes the function defining this command.

        Parameters
        ----------
        context
            The :class:`Context` object representing the current state.
        *args
            Any subcommands of the current command being called.
        **kwargs
            Any other options in the current invocation.
        """

        return self.command(context, *args, **kwargs)

    def dump(self):
        "Returns this command as a dict for registration with the Discord API."
        data = {
            "name": self.name,
            "description": self.description,
            "options": self.options,
            "default_permission": self.default_permission,
        }

        if hasattr(self, "type"):
            data["type"] = self.type

        return data

    def dump_permissions(self):
        return [permission.dump() for permission in self.permissions]

    def autocomplete(self):
        """
        Register an autocomplete handler function for this command.

        Use as a decorator, e.g.

        .. code-block:: python

            @app.command("test")
            def test(ctx, value: str):
                return f"{value}!"

            @test.autocomplete()
            def autocomplete(ctx, value = None):
                return ["hello", "world"]

        """

        def wrapper(f):
            self.discord.add_autocomplete_handler(f, self.name)

        return wrapper


class SlashCommandSubgroup(Command):
    """
    Represents a Subgroup of slash commands.

    Attributes
    ----------
    name
        The name of this subgroup, shown in the Discord client.
    description
        The description of this subgroup, shown in the Discord client.
    is_async
        Whether the subgroup should be considered async (if subcommands
        get an :class:`AsyncContext` instead of a :class:`Context`.)
    """

    def __init__(self, name, description, is_async=False):
        self.name = name
        self.description = description
        self.subcommands = {}
        self.type = ApplicationCommandType.CHAT_INPUT

        self.default_permission = None
        self.permissions = []

        self.is_async = is_async

    def command(self, name=None, description=None, options=None, annotations=None):
        """
        Decorator to create a new Subcommand of this Subgroup.

        Parameters
        ----------
        name
            The name of the command, as displayed in the Discord client.
        description
            The description of the command.
        options
            A list of options for the command, overriding the function's
            keyword arguments.
        annotations
            If ``options`` is not provided, descriptions for each of the
            options defined in the function's keyword arguments.
        """

        def decorator(func):
            nonlocal name, description, options, annotations
            subcommand = Command(func, name, description, options, annotations)
            self.subcommands[subcommand.name] = subcommand
            return func

        return decorator

    @property
    def options(self):
        """
        Returns an array of options that can be passed to this command.
        Computed based on the options of each subcommand or subcommand group.
        """
        options = []
        for command in self.subcommands.values():
            data = command.dump()
            if isinstance(command, SlashCommandSubgroup):
                data["type"] = CommandOptionType.SUB_COMMAND_GROUP
            else:
                data["type"] = CommandOptionType.SUB_COMMAND
            options.append(data)

        return options

    def run(self, context, *subcommands, **kwargs):
        """
        Invokes the relevant subcommand for the given :class:`Context`.

        Parameters
        ----------
        context
            The :class:`Context` object representing the current state.
        *args
            List of subcommands of the current command group being invoked.
        **kwargs
            Any other options in the current invocation.
        """
        return self.subcommands[subcommands[0]].run(context, *subcommands[1:], **kwargs)


class SlashCommandGroup(SlashCommandSubgroup):
    """
    Represents a Subgroup of slash commands.

    Attributes
    ----------
    name
        The name of this subgroup, shown in the Discord client.
    description
        The description of this subgroup, shown in the Discord client.
    is_async
        Whether the subgroup should be considered async (if subcommands
        get an :class:`AsyncContext` instead of a :class:`Context`.)
    default_permission
        Whether the subgroup is enabled by default. Default is True.
    permissions
        List of permission overwrites. These apply to all subcommands of this
        group.
    """

    def __init__(
        self,
        name,
        description,
        is_async=False,
        default_permission=True,
        permissions=None,
    ):
        self.name = name
        self.description = description
        self.subcommands = {}
        self.type = ApplicationCommandType.CHAT_INPUT

        self.default_permission = default_permission
        self.permissions = permissions or []

        self.is_async = is_async

    def subgroup(self, name, description="No description", is_async=False):
        """
        Create a new :class:`SlashCommandSubroup`
        (which can contain multiple subcommands)

        Parameters
        ----------
        name
            The name of the subgroup, as displayed in the Discord client.
        description
            The description of the subgroup. Defaults to "No description".
        is_async
            Whether the subgroup should be considered async (if subcommands
            get an :class:`AsyncContext` instead of a :class:`Context`.)
        """

        group = SlashCommandSubgroup(name, description, is_async)
        self.subcommands[name] = group
        return group



import pytest
import asyncio
import sys

from quart import Quart


@pytest.fixture
def mock_flask():
    # Some ugly hackery required to get Quart's Flask patching to play nice
    # with the actual Flask tests

    del sys.modules["flask"]
    del sys.modules["flask_discord_interactions"]
    del sys.modules["flask_discord_interactions.discord"]

    import quart.flask_patch
    yield None

    del sys.modules["flask"]

    for name in list(sys.modules.keys()):
        if name.startswith("flask."):
            del sys.modules[name]

    del sys.modules["flask_discord_interactions"]
    del sys.modules["flask_discord_interactions.discord"]


@pytest.mark.asyncio
async def test_full_server(mock_flask):
    from flask_discord_interactions import (DiscordInteractions,
                                        InteractionType, ResponseType)

    app = Quart("test_quart")
    app.config["DONT_VALIDATE_SIGNATURE"] = True
    app.config["DONT_REGISTER_WITH_DISCORD"] = True

    discord = DiscordInteractions(app)

    @discord.command()
    async def wait(ctx):
        await asyncio.sleep(0.01)
        return "Hi!"

    discord.set_route_async("/interactions")

    client = app.test_client()

    response = await client.post("/interactions", json={
        "type": InteractionType.APPLICATION_COMMAND,
        "id": 1,
        "channel_id": "",
        "guild_id": "",
        "token": "",
        "data": {
            "id": 1,
            "name": "wait"
        }
    })

    assert response.status_code == 200

    json = await response.get_json()

    assert json["type"] == \
        ResponseType.CHANNEL_MESSAGE_WITH_SOURCE

    assert json["data"]["content"] == "Hi!"



import json

from flask_discord_interactions import Context, Member, Message, ApplicationCommandType

def test_context_parsing():
    # Test data taken from
    # https://discord.com/developers/docs/interactions/application-commands#slash-commands-example-interaction

    data = json.loads("""
        {
            "type": 2,
            "token": "A_UNIQUE_TOKEN",
            "member": {
                "user": {
                    "id": "53908232506183680",
                    "username": "Mason",
                    "avatar": "a_d5efa99b3eeaa7dd43acca82f5692432",
                    "discriminator": "1337",
                    "public_flags": 131141
                },
                "roles": ["539082325061836999"],
                "premium_since": null,
                "permissions": "2147483647",
                "pending": false,
                "nick": null,
                "mute": false,
                "joined_at": "2017-03-13T19:19:14.040000+00:00",
                "is_pending": false,
                "deaf": false
            },
            "id": "786008729715212338",
            "guild_id": "290926798626357999",
            "data": {
                "options": [{
                    "name": "cardname",
                    "value": "The Gitrog Monster"
                }],
                "name": "cardsearch",
                "id": "771825006014889984"
            },
            "channel_id": "645027906669510667"
        }
    """)

    context = Context.from_data(data=data)

    assert context.token == "A_UNIQUE_TOKEN"
    assert context.author.id == "53908232506183680"
    assert context.author.roles == ["539082325061836999"]
    assert context.author.display_name == "Mason"
    assert context.command_name == "cardsearch"
    assert context.options[0]["value"] == "The Gitrog Monster"



def test_author(discord, client):
    @discord.command()
    def display_name(ctx):
        return ctx.author.display_name

    with client.context(Context(author=Member(username="Bob"))):
        assert client.run("display_name").content == "Bob"

    with client.context(Context(author=Member(nick="Dale"))):
        assert client.run("display_name").content == "Dale"


def test_ids(discord, client):
    @discord.command()
    def all_ids(ctx):
        return f"{ctx.author.id}/{ctx.channel_id}/{ctx.guild_id}"

    context = Context(author=Member(id="01"), channel_id="02", guild_id="2003")

    with client.context(context):
        assert client.run("all_ids").content == "01/02/2003"


def test_permissions(discord, client):
    @discord.command()
    def is_admin(ctx):
        return "Yes" if (ctx.author.permissions & 8) else "No"

    context = Context(author=Member(permissions="0"))

    with client.context(context):
        assert client.run("is_admin").content == "No"

    context = Context(author=Member(permissions="8"))

    with client.context(context):
        assert client.run("is_admin").content == "Yes"


def test_user_command_context_parsing():
    # Test data taken from
    # https://discord.com/developers/docs/interactions/application-commands#user-commands-example-interaction

    data = json.loads("""
        {
            "application_id": "775799577604522054",
            "channel_id": "772908445358620702",
            "data": {
                "id": "866818195033292850",
                "name": "context-menu-user-2",
                "resolved": {
                    "members": {
                        "809850198683418695": {
                            "avatar": null,
                            "is_pending": false,
                            "joined_at": "2021-02-12T18:25:07.972000+00:00",
                            "nick": null,
                            "pending": false,
                            "permissions": "246997699136",
                            "premium_since": null,
                            "roles": []
                        }
                    },
                    "users": {
                        "809850198683418695": {
                            "avatar": "afc428077119df8aabbbd84b0dc90c74",
                            "bot": true,
                            "discriminator": "7302",
                            "id": "809850198683418695",
                            "public_flags": 0,
                            "username": "VoltyDemo"
                        }
                    }
                },
                "target_id": "809850198683418695",
                "type": 2
            },
            "guild_id": "772904309264089089",
            "id": "867794291820986368",
            "member": {
                "avatar": null,
                "deaf": false,
                "is_pending": false,
                "joined_at": "2020-11-02T20:46:57.364000+00:00",
                "mute": false,
                "nick": null,
                "pending": false,
                "permissions": "274877906943",
                "premium_since": null,
                "roles": ["785609923542777878"],
                "user": {
                    "avatar": "a_f03401914fb4f3caa9037578ab980920",
                    "discriminator": "6538",
                    "id": "167348773423415296",
                    "public_flags": 1,
                    "username": "ian"
                }
            },
            "token": "UNIQUE_TOKEN",
            "type": 2,
            "version": 1
        }
    """)

    context = Context.from_data(data=data)

    assert context.target.id == "809850198683418695"
    assert context.target.display_name == "VoltyDemo"


def test_message_command_context_parsing():
    # Test data taken from
    # https://discord.com/developers/docs/interactions/application-commands#message-commands-example-interaction

    data = json.loads("""
        {
            "application_id": "775799577604522054",
            "channel_id": "772908445358620702",
            "data": {
                "id": "866818195033292851",
                "name": "context-menu-message-2",
                "resolved": {
                    "messages": {
                        "867793854505943041": {
                            "attachments": [],
                            "author": {
                                "avatar": "a_f03401914fb4f3caa9037578ab980920",
                                "discriminator": "6538",
                                "id": "167348773423415296",
                                "public_flags": 1,
                                "username": "ian"
                            },
                            "channel_id": "772908445358620702",
                            "components": [],
                            "content": "some message",
                            "edited_timestamp": null,
                            "embeds": [],
                            "flags": 0,
                            "id": "867793854505943041",
                            "mention_everyone": false,
                            "mention_roles": [],
                            "mentions": [],
                            "pinned": false,
                            "timestamp": "2021-07-22T15:42:57.744000+00:00",
                            "tts": false,
                            "type": 0
                        }
                    }
                },
                "target_id": "867793854505943041",
                "type": 3
            },
            "guild_id": "772904309264089089",
            "id": "867793873336926249",
            "member": {
                "avatar": null,
                "deaf": false,
                "is_pending": false,
                "joined_at": "2020-11-02T20:46:57.364000+00:00",
                "mute": false,
                "nick": null,
                "pending": false,
                "permissions": "274877906943",
                "premium_since": null,
                "roles": ["785609923542777878"],
                "user": {
                    "avatar": "a_f03401914fb4f3caa9037578ab980920",
                    "discriminator": "6538",
                    "id": "167348773423415296",
                    "public_flags": 1,
                    "username": "ian"
                }
            },
            "token": "UNIQUE_TOKEN",
            "type": 2,
            "version": 1
        }
    """)

    context = Context.from_data(data=data)

    assert context.target.id == "867793854505943041"
    assert context.target.content == "some message"
    assert context.target.timestamp.day == 22
    assert context.target.author.display_name == "ian"


def test_user_command_argument(discord, client):
    @discord.command(type=ApplicationCommandType.MESSAGE)
    def greet(ctx, target):
        return f"Hello, {target.display_name}!"

    with client.context(Context(target=Member(username="Test User"))):
        assert client.run("greet").content == "Hello, Test User!"


def test_message_command_argument(discord, client):
    @discord.command(type=ApplicationCommandType.MESSAGE)
    def repeat(ctx, target):
        return f"I repeat, {target.content.lower()}"

    with client.context(Context(target=Message(content="This is a test."))):
        assert client.run("repeat").content == "I repeat, this is a test."



import dataclasses

from flask_discord_interactions.models.utils import LoadableDataclass



@dataclasses.dataclass
class Role(LoadableDataclass):
    """
    Represents a Role in Discord.

    Attributes
    ----------
    id
        The unique ID (snowflake) of the role.
    name
        The name of the role.
    color
        The color given to the role.
    hoist
        Whether the role is displayed separately in the member list.
    position
        The position of the role in the roles list.
    permissions
        The permissions integer of the role.
    managed
        Whether the role is managed by an integration (bot).
    mentionable
        Whether the role can be mentioned by all users.
    tags
        Miscellaneous information about the role.
    """
    id: str = None
    name: str = None
    color: str = None
    hoist: bool = None
    position: int = None
    managed: bool = None
    mentionable: bool = None
    tags: dict = None


import dataclasses

from flask_discord_interactions.models.utils import LoadableDataclass


@dataclasses.dataclass
class User(LoadableDataclass):
    """
    Represents a User (the identity of a Discord user, not tied to any
    specific guild).

    Attributes
    ----------
    id
        The ID (snowflake) of the user.
    username
        The Discord username of the user.
    discriminator
        The code following the # after the username.
    avatar_hash
        The unique hash identifying the profile picture of the user.
    bot
        Whether the user is a bot account.
    system
        Whether the user is a Discord system account.
    mfa_enabled
        Whether the user has enabled Two-Factor Authentication.
    locale
        The locale of the user.
    flags
        Miscellaneous information about the user.
    premium_type
        The Nitro status of the user.
    public_flags
        Miscellaneous information about the user.
    """
    id: str = None
    username: str = None
    discriminator: str = None
    avatar_hash: str = None
    bot: bool = None
    system: bool = None
    mfa_enabled: bool = None
    locale: str = None
    flags: int = None
    premium_type: int = None
    public_flags: int = None

    @classmethod
    def from_dict(cls, data):
        data = {**data, **data.get("user", {})}
        data["avatar_hash"] = data.get("avatar")
        return super().from_dict(data)

    @property
    def display_name(self):
        "The displayed name of the user (the username)."
        return self.username

    @property
    def avatar_url(self):
        "The URL of the user's profile picture."
        return ("https://cdn.discordapp.com/avatars/"
                f"{self.id}/{self.avatar_hash}.png")


@dataclasses.dataclass
class Member(User):
    """
    Represents a Member (a specific Discord :class:`User` in one particular
    guild.)

    Attributes
    ----------
    nick
        The guild nickname of the user.
    roles
        An array of role IDs that the user has.
    joined_at
        The timestamp that the user joined the guild at.
    premium_since
        The timestamp that the user started Nitro boosting the guild at.
    permissions
        The permissions integer of the user.
    deaf
        Whether the user has been server deafened.
    mute
        Whether the user has been server muted.
    pending
        Whether the user has passed the membership requirements of a guild.
    """
    nick: str = None
    roles: list = None
    joined_at: str = None
    premium_since: str = None
    permissions: int = None
    deaf: bool = None
    mute: bool = None
    pending: bool = None

    def __post_init__(self):
        if isinstance(self.permissions, str):
            self.permissions = int(self.permissions)

    @property
    def display_name(self):
        """
        The displayed name of the user (their nickname, or if none exists,
        their username).
        """
        return self.nick or self.username


from flask_discord_interactions import Message, ActionRow, Button, ButtonStyles


def test_basic_handler(discord, client):
    click_count = 0

    @discord.custom_handler()
    def handle_click(ctx):
        nonlocal click_count
        click_count += 1

        return Message(
            content=f"The button has been clicked {click_count} times",
            components=[
                ActionRow(components=[
                    Button(
                        style=ButtonStyles.PRIMARY,
                        custom_id=handle_click,
                        label="Click Me!"
                    )
                ])
            ],
            update=True
        )


    # The main command sends the initial Response
    @discord.command()
    def click_counter(ctx):
        "Count the number of button clicks"

        return Message(
            content=f"The button has been clicked {click_count} times",
            components=[
                ActionRow(components=[
                    Button(
                        style=ButtonStyles.PRIMARY,
                        custom_id=handle_click,
                        label="Click Me!"
                    )
                ])
            ]
        )

    client.run("click_counter")
    discord.custom_id_handlers[handle_click](None)
    assert click_count == 1


def test_stateful_handler(discord, client):
    @discord.custom_handler()
    def handle_click(ctx, click_count):
        click_count = int(click_count)
        click_count += 1

        return Message(
            content=f"{click_count} clicks",
            components=[
                ActionRow(components=[
                    Button(
                        style=ButtonStyles.PRIMARY,
                        custom_id=[handle_click, click_count],
                        label="Click Me!"
                    )
                ])
            ],
            update=True
        )


    # The main command sends the initial Response
    @discord.command()
    def click_counter(ctx):
        "Count the number of button clicks"

        return Message(
            content=f"The button has been clicked 0 times",
            components=[
                ActionRow(components=[
                    Button(
                        style=ButtonStyles.PRIMARY,
                        custom_id=[handle_click, 0],
                        label="Click Me!"
                    )
                ])
            ]
        )

    client.run("click_counter")
    response = discord.custom_id_handlers[handle_click](None, 0)
    assert response.content == "1 clicks"


import copy
from contextlib import contextmanager
from flask_discord_interactions.models.command import ApplicationCommandType

from flask_discord_interactions.models import Message
from flask_discord_interactions.context import Context
from flask_discord_interactions.command import SlashCommandSubgroup


class Client:
    """
    A class to represent a mock client that can be used to execute
    Application Commands programatically without connecting to Discord.

    Attributes
    ----------
    discord
        The DiscordInteractions object to connect to.
    """
    def __init__(self, discord):
        self.discord = discord
        self.current_context = Context()

    @contextmanager
    def context(self, context=None, **kwargs):
        """
        Uses the specified context object for commands run. This is a context
        manager meant to be used with the ``with`` statement.

        Parameters
        ----------
        context
            The :class:`.Context` object to use.
        **kwargs
            Keyword arguments to use to construct the :class:`.Context` object.
        """
        if context:
            self.current_context = context
        elif kwargs:
            self.current_context = Context(**kwargs)
        else:
            self.current_context = Context()
        try:
            yield self.current_context
        finally:
            self.current_context = Context()

    def run(self, *names, **params):
        """
        Run a specified Application Command.

        Parameters
        ----------
        *names
            The names of the command, subcommand group, and subcommand (if
            present). This may contain just one name (in the case of a
            root-level command), or up to three (for a subcommand inside a
            subcommand group).
        **params
            Options to pass to the command being called.
        """
        command = self.discord.discord_commands[names[0]]

        if command.type == ApplicationCommandType.CHAT_INPUT:
            i = 1
            for i in range(1, len(names)):
                if not isinstance(command, SlashCommandSubgroup):
                    break
                command = command.subcommands[names[i]]
            else:
                i += 1

            return Message.from_return_value(
                command.run(self.current_context, *names[i:], **params))

        return Message.from_return_value(
            command.run(self.current_context, self.current_context.target))

    def run_handler(self, custom_id, *args):
        """
        Run a specified custom ID handler.

        Parameters
        ----------
        custom_id
            The ID of the handler function to run.
        *args
            Options to pass to the handler being called.
        """

        new_context = copy.deepcopy(self.current_context)

        new_context.custom_id = "\n".join((custom_id, *args))
        new_context.parse_custom_id()

        handler = self.discord.custom_id_handlers[new_context.primary_id]

        args = new_context.create_handler_args(handler)

        response = handler(new_context, *args)
        return Message.from_return_value(response)



from dataclasses import dataclass
from typing import Any, List, Union
import inspect
import itertools
import warnings
import types

import requests

from flask_discord_interactions.models import (
    LoadableDataclass,
    Member,
    Channel,
    Role,
    User,
    CommandOptionType,
    ApplicationCommandType,
    Message,
)
from flask_discord_interactions.models.option import Option


@dataclass
class Context(LoadableDataclass):
    """
    Represents the context in which a :class:`Command` or custom ID
    handler is invoked.

    Attributes
    ----------
    author
        A :class:`User` or  :class:`Member` object representing the invoking
        user.
    id
        The unique ID (snowflake) of this interaction.
    type
        The :class:`ApplicationCommandType` of this interaction.
    token
        The token to use when sending followup messages.
    channel_id
        The unique ID (snowflake) of the channel this command was invoked in.
    guild_id
        The unique ID (snowflake) of the guild this command was invoked in.
    options
        A list of the options passed to the command.
    values
        A list of the values selected, if this is a Select Menu handler.
    resolved
        Additional data ()
    command_name
        The name of the command that was invoked.
    command_id
        The unique ID (snowflake) of the command that was invoked.
    users
        :class:`User` objects for each user specified as an option.
    members
        :class:`Member` objects for each user specified as an option, if this
        command was invoked in a guild.
    channels
        :class:`Channel` objects for each channel specified as an option.
    roles
        :class:`Role` object for each role specified as an option.
    target
        The targeted :class:`User` or message.
    message
        The message that the invoked components are attached to. Only available on component interactions
    """

    author: Union[Member, User] = None
    id: str = None
    type: int = None
    token: str = None
    channel_id: str = None
    guild_id: str = None
    options: list = None
    values: list = None
    resolved: dict = None
    command_name: str = None
    command_id: str = None
    members: List[Member] = None
    channels: List[Channel] = None
    roles: List[Role] = None
    message: Message = None

    app: Any = None
    discord: Any = None

    custom_id: str = None
    primary_id: str = None
    handler_state: list = None

    target_id: str = None
    target: Union[User, Message] = None

    @classmethod
    def from_data(cls, discord=None, app=None, data={}):
        if data is None:
            data = {}

        # If this is a proxy (e.g. flask.current_app), get the current object
        # https://flask.palletsprojects.com/en/2.0.x/reqcontext/#notes-on-proxies
        if hasattr(app, "_get_current_object"):
            app = app._get_current_object()

        result = cls(
            app=app,
            discord=discord,
            id=data.get("id"),
            type=data.get("data", {}).get("type") or ApplicationCommandType.CHAT_INPUT,
            token=data.get("token"),
            channel_id=data.get("channel_id"),
            guild_id=data.get("guild_id"),
            options=data.get("data", {}).get("options"),
            values=data.get("data", {}).get("values", []),
            resolved=data.get("data", {}).get("resolved", {}),
            command_name=data.get("data", {}).get("name"),
            command_id=data.get("data", {}).get("id"),
            custom_id=data.get("data", {}).get("custom_id") or "",
            target_id=data.get("data", {}).get("target_id"),
        )

        result.data = data

        result.parse_author(data)
        result.parse_message(data)
        result.parse_custom_id()
        result.parse_resolved()
        result.parse_target()
        return result

    @property
    def auth_headers(self):
        if self.discord:
            return self.discord.auth_headers(self.app)
        else:
            return self.frozen_auth_headers

    def parse_author(self, data):
        """
        Parse the author (invoking user) of this interaction.

        This will set :attr:`author` to a :class:`User` object if this
        interaction occurred in a direct message, or a :class:`Member` object
        if interaction occurred in a guild.

        Parameters
        ----------
        data
            The incoming interaction data.
        """
        if data.get("member"):
            self.author = Member.from_dict(data["member"])
        elif data.get("user"):
            self.author = User.from_dict(data["user"])
        else:
            self.author = None

    def parse_message(self, data):
        """
        Parse the message out of in interaction.

        Parameters
        ----------
        data
            The incoming interaction data.
        """
        if data.get("message"):
            self.message = Message.from_dict(data["message"])
        else:
            self.message = None

    def parse_custom_id(self):
        """
        Parse the custom ID of the incoming interaction data.

        This includes the primary ID as well as any state stored in the
        handler.
        """

        self.primary_id = self.custom_id.split("\n", 1)[0]
        self.handler_state = self.custom_id.split("\n")

    def parse_resolved(self):
        """
        Parse the ``"resolved"`` section of the incoming interaction data.

        This section includes objects representing each user, member, channel,
        and role passed as an argument to the command.
        """

        self.members = {}
        for id in self.resolved.get("members", {}):
            member_info = self.resolved["members"][id]
            member_info["user"] = self.resolved["users"][id]
            self.members[id] = Member.from_dict(member_info)

        self.users = {
            id: User.from_dict(data)
            for id, data in self.resolved.get("users", {}).items()
        }

        self.channels = {
            id: Channel.from_dict(data)
            for id, data in self.resolved.get("channels", {}).items()
        }

        self.roles = {
            id: Role.from_dict(data)
            for id, data in self.resolved.get("roles", {}).items()
        }

        self.messages = {
            id: Message.from_dict(data)
            for id, data in self.resolved.get("messages", {}).items()
        }

    def parse_target(self):
        """
        Parse the target of the incoming interaction.

        For User and Message commands, the target is the relevant user or
        message. This method sets the `ctx.target` field.
        """
        if self.type == ApplicationCommandType.USER:
            self.target = self.members[self.target_id]
        elif self.type == ApplicationCommandType.MESSAGE:
            self.target = self.messages[self.target_id]
        else:
            self.target = None

    def create_args(self):
        """
        Create the arguments which will be passed to the function when the
        :class:`Command` is invoked.
        """
        if self.type == ApplicationCommandType.CHAT_INPUT:
            return self.create_args_chat_input()
        elif self.type == ApplicationCommandType.USER:
            return [self.target], {}
        elif self.type == ApplicationCommandType.MESSAGE:
            return [self.target], {}

    def create_args_chat_input(self):
        """
        Create the arguments for this command, assuming it is a ``CHAT_INPUT``
        command.
        """

        def create_args_recursive(data, resolved):
            if not data.get("options"):
                return [], {}

            args = []
            kwargs = {}

            for option in data["options"]:
                if option["type"] in [
                    CommandOptionType.SUB_COMMAND,
                    CommandOptionType.SUB_COMMAND_GROUP,
                ]:

                    args.append(option["name"])

                    sub_args, sub_kwargs = create_args_recursive(option, resolved)

                    args += sub_args
                    kwargs.update(sub_kwargs)

                elif option["type"] == CommandOptionType.USER:
                    if "members" in resolved:
                        member_data = resolved["members"][option["value"]]
                        member_data["user"] = resolved["users"][option["value"]]

                        kwargs[option["name"]] = Member.from_dict(member_data)
                    else:
                        kwargs[option["name"]] = User.from_dict(
                            resolved["users"][option["value"]]
                        )

                elif option["type"] == CommandOptionType.CHANNEL:
                    kwargs[option["name"]] = Channel.from_dict(
                        resolved["channels"][option["value"]]
                    )

                elif option["type"] == CommandOptionType.ROLE:
                    kwargs[option["name"]] = Role.from_dict(
                        resolved["roles"][option["value"]]
                    )

                else:
                    kwargs[option["name"]] = option["value"]

            return args, kwargs

        return create_args_recursive({"options": self.options}, self.resolved)

    def create_handler_args(self, handler):
        """
        Create the arguments which will be passed to the function when a
        custom ID handler is invoked.

        Parameters
        ----------
        handler
            The custom ID handler to create arguments for.
        """

        args = self.handler_state[1:]

        sig = inspect.signature(handler)

        iterator = zip(
            itertools.count(), args, itertools.islice(sig.parameters.values(), 1, None)
        )

        for i, argument, parameter in iterator:
            annotation = parameter.annotation

            if annotation == int:
                args[i] = int(argument)

            elif annotation == bool:
                if argument == "True":
                    args[i] = True
                elif argument == "False":
                    args[i] = False
                elif argument == "None":
                    args[i] = None
                else:
                    raise ValueError(
                        f"Invalid bool in handler state parsing: {args[i]}"
                    )

        return args

    def create_autocomplete_args(self):
        return [Option.from_data(option) for option in self.options]

    def followup_url(self, message=None):
        """
        Return the followup URL for this interaction. This URL can be used to
        send a new message, or to edit or delete an existing message.

        Parameters
        ----------
        message
            The ID of the message to edit or delete.
            If None, sends a new message.
            If "@original", refers to the original message.
        """

        url = (
            f"{self.app.config['DISCORD_BASE_URL']}/webhooks/"
            f"{self.app.config['DISCORD_CLIENT_ID']}/{self.token}"
        )
        if message is not None:
            url += f"/messages/{message}"

        return url

    def edit(self, updated, message="@original"):
        """
        Edit an existing message.

        Parameters
        ----------
        updated
            The updated Message to edit the message to.
        message
            The ID of the message to edit.
            If omitted, edits the original message.
        """

        updated = Message.from_return_value(updated)

        if not self.app or self.app.config["DONT_REGISTER_WITH_DISCORD"]:
            return

        updated = requests.patch(
            self.followup_url(message),
            json=updated.dump_followup(),
            headers=self.auth_headers,
        )
        updated.raise_for_status()

    def delete(self, message="@original"):
        """
        Delete an existing message.

        Parameters
        ----------
        message
            The ID of the message to delete.
            If omitted, deletes the original message.
        """

        if not self.app or self.app.config["DONT_REGISTER_WITH_DISCORD"]:
            return

        response = requests.delete(
            self.followup_url(message), headers=self.auth_headers
        )
        response.raise_for_status()

    def send(self, message):
        """
        Send a new followup message.

        Parameters
        ----------
        message
            The :class:`Message` to send as a followup message.
        """

        if not self.app or self.app.config["DONT_REGISTER_WITH_DISCORD"]:
            return

        message = Message.from_return_value(message)

        message = requests.post(
            self.followup_url(), headers=self.auth_headers, **message.dump_multipart()
        )
        message.raise_for_status()
        return message.json()["id"]

    def get_command(self, command_name=None):
        "Get the ID of a command by name."
        if command_name is None:
            return self.command_id
        else:
            try:
                return self.app.discord_commands[command_name].id
            except KeyError:
                raise ValueError(f"Unknown command: {command_name}")

    def overwrite_permissions(self, permissions, command=None):
        """
        Overwrite the permission overwrites for this command.

        Parameters
        ----------
        permissions
            The new list of permission overwrites.
        command
            The name of the command to overwrite permissions for. If omitted,
            overwrites for the invoking command.
        """

        url = (
            f"{self.app.config['DISCORD_BASE_URL']}/"
            f"applications/{self.app.config['DISCORD_CLIENT_ID']}/"
            f"guilds/{self.guild_id}/"
            f"commands/{self.get_command(command)}/permissions"
        )

        data = [permission.dump() for permission in permissions]

        if self.app.config["DONT_REGISTER_WITH_DISCORD"]:
            return

        response = requests.put(
            url, headers=self.auth_headers, json={"permissions": data}
        )
        response.raise_for_status()

    def freeze(self):
        "Return a copy of this Context that can be pickled for RQ and Celery."

        app = types.SimpleNamespace()

        CONFIG_KEYS = [
            "DISCORD_BASE_URL",
            "DISCORD_CLIENT_ID",
            "DONT_REGISTER_WITH_DISCORD",
        ]

        app.config = {key: self.app.config[key] for key in CONFIG_KEYS}

        new_context = Context.from_data(app=app, data=self.data)
        new_context.frozen_auth_headers = self.auth_headers

        return new_context


@dataclass
class AsyncContext(Context):
    """
    Represents the context in which an asynchronous :class:`Command` is
    invoked. Also provides coroutine functions to handle followup messages.

    Users should not need to instantiate this class manually.
    """

    def __post_init__(self):
        if not self.app or self.app.config["DONT_REGISTER_WITH_DISCORD"]:
            return

        self.session = self.app.discord_client_session

    async def edit(self, updated, message="@original"):
        """
        Edit an existing message.

        Parameters
        ----------
        updated
            The updated Message to edit the message to.
        message
            The ID of the message to edit.
            If omitted, edits the original message.
        """

        updated = Message.from_return_value(updated)

        if not self.app or self.app.config["DONT_REGISTER_WITH_DISCORD"]:
            return

        await self.session.patch(
            self.followup_url(message), json=updated.dump_followup()
        )

    async def delete(self, message="@original"):
        """
        Delete an existing message.

        Parameters
        ----------
        message
            The ID of the message to delete.
            If omitted, deletes the original message.
        """

        if not self.app or self.app.config["DONT_REGISTER_WITH_DISCORD"]:
            return

        await self.session.delete(self.followup_url(message))

    async def send(self, message):
        """
        Send a new followup message.

        Parameters
        ----------
        message
            The Message object to send as a followup message.
        """

        message = Message.from_return_value(message)

        if not self.app or self.app.config["DONT_REGISTER_WITH_DISCORD"]:
            return

        async with self.session.post(
            self.followup_url(), headers=self.auth_headers, **message.dump_multipart()
        ) as message:
            return (await message.json())["id"]

    async def overwrite_permissions(self, permissions, command=None):
        """
        Overwrite the permission overwrites for this command.

        Parameters
        ----------
        permissions
            The new list of permission overwrites.
        command
            The name of the command to overwrite permissions for. If omitted,
            overwrites for the invoking command.
        """

        url = (
            f"{self.app.config['DISCORD_BASE_URL']}/"
            f"applications/{self.app.config['DISCORD_CLIENT_ID']}/"
            f"guilds/{self.guild_id}/"
            f"commands/{self.get_command(command)}/permissions"
        )

        data = [permission.dump() for permission in permissions]

        if not self.app or self.app.config["DONT_REGISTER_WITH_DISCORD"]:
            return

        await self.session.put(
            url, headers=self.auth_headers, json={"permissions": data}
        )

    async def close(self):
        """
        Deprecated as of v1.0.2.

        Previously, this closed the AsyncContext's aiohttp ClientSession that
        was used to send followup messages. This is no longer necessary, as
        this library now maintains a single ClientSession for the entire
        application.
        """

        warnings.warn(
            "Deprecated! AsyncContext.close is a no-op. "
            "Since v1.0.2, only one aiohttp ClientSession is created "
            "for all requests to Discord for the app. "
            "Thus, there is no need to close the AsyncContext. ",
            DeprecationWarning,
        )



class Permission:
    """
    An object representing a single permission overwrite.

    ``Permission(role='1234')`` allows users with role ID 1234 to use the
    command

    ``Permission(user='5678')`` allows user ID 5678 to use the command

    ``Permission(role='9012', allow=False)`` denies users with role ID 9012
    from using the command
    """

    def __init__(self, role=None, user=None, allow=True):
        if bool(role) == bool(user):
            raise ValueError("specify only one of role or user")

        self.type = 1 if role else 2
        self.id = role or user
        self.permission = allow

    def dump(self):
        "Returns a dict representation of the permission"
        return {"type": self.type, "id": self.id, "permission": self.permission}



import enum

from flask_discord_interactions import Member, Channel, Role


def test_str(discord, client):
    @discord.command()
    def repeat(ctx, message: str):
        return message + " " + message

    assert client.run("repeat", message="meow").content == "meow meow"


def test_int(discord, client):
    @discord.command()
    def add_one(ctx, number: int):
        return str(number + 1)

    assert client.run("add_one", number=2).content == "3"


def test_bool(discord, client):
    @discord.command()
    def and_gate(ctx, a: bool, b: bool):
        return str(a and b)

    assert client.run("and_gate", a=True, b=False).content == "False"
    assert client.run("and_gate", a=True, b=True).content == "True"


def test_number(discord, client):
    @discord.command()
    def round_to_nearest(ctx, number: float):
        return str(round(number))

    assert client.run("round_to_nearest", number=1.5).content == "2"
    assert client.run("round_to_nearest", number=1.25).content == "1"


def test_default(discord, client):
    @discord.command()
    def square(ctx, number: int = 5):
        return str(number ** 2)

    assert client.run("square").content == "25"
    assert client.run("square", number=2).content == "4"


def test_falsy_or_none(discord, client):
    @discord.command()
    def falsy_or_none(ctx, arg: bool = None):
        if arg is None:
            return "None"
        elif arg:
            return "True"
        else:
            return "False"

    assert client.run("falsy_or_none").content == "None"
    assert client.run("falsy_or_none", arg=False).content == "False"
    assert client.run("falsy_or_none", arg=True).content == "True"


def test_str_enum(discord, client):
    class Animal(enum.Enum):
        Dog = "dog"
        Cat = "cat"


    @discord.command()
    def favorite(ctx, choice: Animal):
        return f"You chose {choice}!"

    assert client.run("favorite", choice="dog").content == "You chose dog!"


def test_int_enum(discord, client):
    class BigNumber(enum.IntEnum):
        thousand = 1_000
        million = 1_000_000
        billion = 1_000_000_000
        trillion = 1_000_000_000_000


    @discord.command()
    def big_number(ctx, number: BigNumber):
        return f"One more is {number+1}"

    assert client.run("big_number", number=1_000).content == "One more is 1001"


def test_role(discord, client):
    @discord.command()
    def role_id(ctx, role: Role):
        return role.id

    assert client.run("role_id", role=Role(id="1234")).content == "1234"


def test_channel(discord, client):
    @discord.command()
    def channel_name(ctx, channel: Channel):
        return channel.name

    assert client.run(
        "channel_name", channel=Channel(name="general")).content == "general"


def test_member(discord, client):
    @discord.command()
    def ship_them(ctx, other: Member):
        return f"{ctx.author.display_name} <3 {other.display_name}"

    with client.context(author=Member(username="Romeo")):
        assert client.run(
            "ship_them", other=Member(username="Juliet")).content == \
                "Romeo <3 Juliet"

        assert client.run(
            "ship_them",
            other=Member(username="Juliet", nick="J")).content == \
                "Romeo <3 J"


def test_multiple(discord, client):
    @discord.command()
    def has_role(ctx, role: Role, user: Member = None):
        if user is None:
            user = ctx.author

        return str(role.id in user.roles)

    with client.context(author=Member(roles=["2", "4", "8", "16"])):
        assert client.run("has_role", role=Role(id="2")).content == "True"
        assert client.run("has_role", role=Role(id="7")).content == "False"

        user = Member(roles=["7"])
        assert client.run(
            "has_role", user=user, role=Role(id="2")).content == "False"
        assert client.run(
            "has_role", user=user, role=Role(id="7")).content == "True"



import time
import inspect
import uuid
import atexit
import asyncio
import warnings

import requests

from flask import current_app, request, jsonify, abort

from nacl.exceptions import BadSignatureError
from nacl.signing import VerifyKey

from flask_discord_interactions.models.autocomplete import AutocompleteResult

try:
    import aiohttp
except ImportError:
    aiohttp = None

from flask_discord_interactions.command import Command, SlashCommandGroup
from flask_discord_interactions.context import Context, ApplicationCommandType
from flask_discord_interactions.models import Message, ResponseType


class InteractionType:
    PING = 1
    APPLICATION_COMMAND = 2
    MESSAGE_COMPONENT = 3
    APPLICATION_COMMAND_AUTOCOMPLETE = 4
    MODAL_SUBMIT = 5


class DiscordInteractionsBlueprint:
    """
    Represents a collection of :class:`ApplicationCommand` s.

    Useful for splitting a bot across multiple files.
    """

    def __init__(self):
        self.discord_commands = {}
        self.custom_id_handlers = {}
        self.autocomplete_handlers = {}

    def add_command(
        self,
        command,
        name=None,
        description=None,
        options=None,
        annotations=None,
        type=ApplicationCommandType.CHAT_INPUT,
        default_permission=None,
        permissions=None,
    ):
        """
        Create and add a new :class:`ApplicationCommand`.

        Parameters
        ----------
        command
            Function to execute when the command is run.
        name
            The name of the command, as displayed in the Discord client.
        description
            The description of the command.
        options
            A list of options for the command, overriding the function's
            keyword arguments.
        annotations
            If ``options`` is not provided, descriptions for each of the
            options defined in the function's keyword arguments.
        type
            The ``ApplicationCommandType`` of the command.
        default_permission
            Whether the command is enabled by default. Default is True.
        permissions
            List of permission overwrites.
        """
        command = Command(
            command,
            name,
            description,
            options,
            annotations,
            type,
            default_permission,
            permissions,
            self,
        )
        self.discord_commands[command.name] = command
        return command

    def add_slash_command(self, *args, **kwargs):
        """
        Deprecated! As of v1.1.0, ``add_slash_command`` has been renamed to
        :meth:`add_command`, as it can now add User and Message commands.
        """
        warnings.warn(
            "Deprecated! As of v1.1.0, add_slash_command has been renamed to "
            "add_command, as it can now add User and Message commands.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.add_command(*args, **kwargs)

    def command(
        self,
        name=None,
        description=None,
        options=None,
        annotations=None,
        type=ApplicationCommandType.CHAT_INPUT,
        default_permission=None,
        permissions=None,
    ):
        """
        Decorator to create a new :class:`Command`.

        Parameters
        ----------
        name
            The name of the command, as displayed in the Discord client.
        description
            The description of the command.
        options
            A list of options for the command, overriding the function's
            keyword arguments.
        annotations
            If ``options`` is not provided, descriptions for each of the
            options defined in the function's keyword arguments.
        type
            The ``ApplicationCommandType`` of the command.
        default_permission
            Whether the command is enabled by default. Default is True.
        permissions
            List of permission overwrites.
        """

        def decorator(func):
            nonlocal name, description, type, options
            command = self.add_command(
                func,
                name,
                description,
                options,
                annotations,
                type,
                default_permission,
                permissions,
            )
            return command

        return decorator

    def command_group(
        self,
        name,
        description="No description",
        is_async=False,
        default_permission=None,
        permissions=None,
    ):
        """
        Create a new :class:`SlashCommandGroup`
        (which can contain multiple subcommands)

        Parameters
        ----------
        name
            The name of the command group, as displayed in the Discord client.
        description
            The description of the command group.
        is_async
            Whether the subgroup should be considered async (if subcommands
            get an :class:`.AsyncContext` instead of a :class:`Context`.)
        default_permission
            Whether the command group is enabled by default.
        permissions
            List of permission overwrites. These apply to the entire group.
        """

        group = SlashCommandGroup(
            name, description, is_async, default_permission, permissions
        )
        self.discord_commands[name] = group
        return group

    def add_custom_handler(self, handler, custom_id=None):
        """
        Add a handler for an incoming interaction with the specified custom ID.

        Parameters
        ----------
        handler
            The function to call to handle the incoming interaction.
        custom_id
            The custom ID to respond to. If not specified, the ID will be
            generated randomly.

        Returns
        -------
        str
            The custom ID that the handler will respond to.
        """
        if custom_id is None:
            custom_id = str(uuid.uuid4())

        self.custom_id_handlers[custom_id] = handler
        return custom_id

    def custom_handler(self, custom_id=None):
        """
        Returns a decorator to register a handler for a custom ID.

        Parameters
        ----------
        custom_id
            The custom ID to respond to. If not specified, the ID will be
            generated randomly.
        """

        def decorator(func):
            nonlocal custom_id
            custom_id = self.add_custom_handler(func, custom_id)
            return custom_id

        return decorator

    def add_autocomplete_handler(self, handler, command_name):
        """
        Add a handler for an incoming autocomplete request.

        Parameters
        ----------
        handler
            The function to call to handle the incoming autocomplete request.
        command_name
            The name of the command to autocomplete.
        """
        self.autocomplete_handlers[command_name] = handler


class DiscordInteractions(DiscordInteractionsBlueprint):
    """
    Handles registering a collection of :class:`Command` s, receiving
    incoming interaction data, and sending/editing/deleting messages via
    webhook.
    """

    def __init__(self, app=None):
        super().__init__()

        self.app = app
        if app is not None:
            self.init_app(app)

    def init_app(self, app):
        """
        Initialize a Flask app with Discord-specific configuration and
        attributes.

        Parameters
        ----------
        app
            The Flask app to initialize.
        """

        app.config.setdefault("DISCORD_BASE_URL", "https://discord.com/api/v9")
        app.config.setdefault("DISCORD_CLIENT_ID", "")
        app.config.setdefault("DISCORD_PUBLIC_KEY", "")
        app.config.setdefault("DISCORD_CLIENT_SECRET", "")
        app.config.setdefault("DONT_VALIDATE_SIGNATURE", False)
        app.config.setdefault("DONT_REGISTER_WITH_DISCORD", False)
        app.discord_commands = self.discord_commands
        app.custom_id_handlers = self.custom_id_handlers
        app.autocomplete_handlers = self.autocomplete_handlers
        app.discord_token = None

    def fetch_token(self, app=None):
        """
        Fetch an OAuth2 token from Discord using the ``CLIENT_ID`` and
        ``CLIENT_SECRET`` with the ``applications.commands.update`` scope. This
        can be used to register new application commands.

        Parameters
        ----------
        app
            The Flask app with the relevant config (client ID and secret).
        """

        if app is None:
            app = self.app

        if app.config["DONT_REGISTER_WITH_DISCORD"]:
            app.discord_token = {
                "token_type": "Bearer",
                "scope": "applications.commands.update",
                "expires_in": 604800,
                "access_token": "DONT_REGISTER_WITH_DISCORD",
            }
            app.discord_token["expires_on"] = (
                time.time() + app.discord_token["expires_in"] / 2
            )
            return
        response = requests.post(
            app.config["DISCORD_BASE_URL"] + "/oauth2/token",
            data={
                "grant_type": "client_credentials",
                "scope": "applications.commands.update",
            },
            headers={"Content-Type": "application/x-www-form-urlencoded"},
            auth=(app.config["DISCORD_CLIENT_ID"], app.config["DISCORD_CLIENT_SECRET"]),
        )

        response.raise_for_status()
        app.discord_token = response.json()
        app.discord_token["expires_on"] = (
            time.time() + app.discord_token["expires_in"] / 2
        )

    def auth_headers(self, app):
        """
        Get the Authorization header required for HTTP requests to the
        Discord API.

        Parameters
        ----------
        app
            The Flask app with the relevant access token.
        """

        if app.discord_token is None or time.time() > app.discord_token["expires_on"]:
            self.fetch_token(app)
        return {"Authorization": f"Bearer {app.discord_token['access_token']}"}

    def update_commands(self, app=None, guild_id=None):
        """
        Update the list of commands registered with Discord.
        This method will overwrite all existing commands.

        Make sure you aren't calling this every time a new worker starts! You
        will run into rate-limiting issues if multiple workers attempt to
        register commands simultaneously. Read :ref:`workers` for more
        info.

        Parameters
        ----------
        app
            The Flask app with the relevant Discord access token.
        guild_id
            The ID of the Discord guild to register commands to. If omitted,
            the commands are registered globally.
        """

        if app is None:
            app = self.app

        if guild_id:
            url = (
                f"{app.config['DISCORD_BASE_URL']}/applications/"
                f"{app.config['DISCORD_CLIENT_ID']}/"
                f"guilds/{guild_id}/commands"
            )
        else:
            url = (
                f"{app.config['DISCORD_BASE_URL']}/applications/"
                f"{app.config['DISCORD_CLIENT_ID']}/commands"
            )

        overwrite_data = [command.dump() for command in app.discord_commands.values()]

        if not app.config["DONT_REGISTER_WITH_DISCORD"]:
            response = requests.put(
                url, json=overwrite_data, headers=self.auth_headers(app)
            )
            self.throttle(response)

            try:
                response.raise_for_status()
            except requests.exceptions.HTTPError:
                raise ValueError(
                    f"Unable to register commands:"
                    f"{response.status_code} {response.text}"
                )

            for command in response.json():
                if command["name"] in app.discord_commands:
                    app.discord_commands[command["name"]].id = command["id"]
        else:
            for command in app.discord_commands.values():
                command.id = command.name

        url += "/permissions"

        permissions_data = [
            {"id": command.id, "permissions": command.dump_permissions()}
            for command in app.discord_commands.values()
            if command.permissions
        ]

        if not permissions_data:
            return

        if not app.config["DONT_REGISTER_WITH_DISCORD"]:
            response = requests.put(
                url, json=permissions_data, headers=self.auth_headers(app)
            )
            self.throttle(response)

            try:
                response.raise_for_status()
            except requests.exceptions.HTTPError:
                raise ValueError(
                    f"Unable to register permissions:"
                    f"{response.status_code} {response.text}"
                )

    def update_slash_commands(self, *args, **kwargs):
        """
        Deprecated! As of v1.1.0, ``update_slash_commands`` has been renamed to
        ``update_commands``, as it updates User and Message commands as well.
        """
        warnings.warn(
            "Deprecated! As of v1.1.0, update_slash_commands has been renamed "
            "to update_commands, as it updates User and Message commands too.",
            DeprecationWarning,
            stacklevel=2,
        )

        return self.update_commands(*args, **kwargs)

    def throttle(self, response):
        """
        Throttle the number of HTTP requests made to Discord
        using the ``X-RateLimit`` headers
        https://discord.com/developers/docs/topics/rate-limits

        Parameters
        ----------
        response
            Response object from a previous HTTP request
        """

        rate_limit_remaining = int(response.headers["X-RateLimit-Remaining"])
        rate_limit_reset = float(response.headers["X-RateLimit-Reset"])
        # rate_limit_limit = response.headers["X-RateLimit-Limit"]
        # rate_limit_bucket = response.headers["X-RateLimit-Bucket"]

        if not rate_limit_remaining:
            time.sleep(max(rate_limit_reset - time.time(), 0))

    def register_blueprint(self, blueprint, app=None):
        """
        Register a :class:`DiscordInteractionsBlueprint` to this
        DiscordInteractions class. Updates this instance's list of
        :class:`Command` s using the blueprint's list of
        :class:`Command` s.

        Parameters
        ----------
        blueprint
            The :class:`DiscordInteractionsBlueprint` to add
            :class:`Command` s from.
        app
            The Flask app with the relevant Discord commands.
        """

        if app is None:
            app = self.app

        app.discord_commands.update(blueprint.discord_commands)
        app.custom_id_handlers.update(blueprint.custom_id_handlers)
        app.autocomplete_handlers.update(blueprint.autocomplete_handlers)

    def run_command(self, data):
        """
        Run the corresponding :class:`Command` given incoming interaction
        data.

        Parameters
        ----------
        data
            Incoming interaction data.
        """

        command_name = data["data"]["name"]

        command = current_app.discord_commands.get(command_name)

        if command is None:
            raise ValueError(f"Invalid command name: {command_name}")

        return command.make_context_and_run(self, current_app, data)

    def run_handler(self, data):
        """
        Run the corresponding custom ID handler given incoming interaction
        data.

        Parameters
        ----------
        data
            Incoming interaction data.
        """

        context = Context.from_data(self, current_app, data)
        handler = self.custom_id_handlers[context.primary_id]
        args = context.create_handler_args(handler)
        result = handler(context, *args)

        return Message.from_return_value(result)

    def run_autocomplete(self, data):
        """
        Run the corresponding autocomplete handler given incoming interaction
        data.

        Parameters
        ----------
        data
            Incoming interaction data.
        """

        context = Context.from_data(self, current_app, data)
        handler = self.autocomplete_handlers[context.command_name]
        args = context.create_autocomplete_args()
        result = handler(context, *args)

        return AutocompleteResult.from_return_value(result)

    def verify_signature(self, request):
        """
        Verify the signature sent by Discord with incoming interactions.

        Parameters
        ----------
        request
            The request to verify the signature of.
        """

        signature = request.headers.get("X-Signature-Ed25519")
        timestamp = request.headers.get("X-Signature-Timestamp")

        if current_app.config["DONT_VALIDATE_SIGNATURE"]:
            return

        if signature is None or timestamp is None:
            abort(401, "Missing signature or timestamp")

        message = timestamp.encode() + request.data
        verify_key = VerifyKey(bytes.fromhex(current_app.config["DISCORD_PUBLIC_KEY"]))
        try:
            verify_key.verify(message, bytes.fromhex(signature))
        except BadSignatureError:
            abort(401, "Incorrect Signature")

        if not request.json:
            abort(400, "Request JSON required")

    def set_route(self, route, app=None):
        """
        Add a route handler to the Flask app that handles incoming
        interaction data.

        If you are using Quart, you should use
        :meth:`.DiscordInteractions.set_route_async` instead.

        Parameters
        ----------
        route
            The URL path to receive interactions on.
        app
            The Flask app to add the route to.
        """

        if app is None:
            app = self.app

        @app.route(route, methods=["POST"])
        def interactions():
            self.verify_signature(request)

            interaction_type = request.json.get("type")
            if interaction_type == InteractionType.PING:
                return jsonify({"type": ResponseType.PONG})

            elif interaction_type == InteractionType.APPLICATION_COMMAND:
                return jsonify(self.run_command(request.json).dump())

            elif interaction_type == InteractionType.MESSAGE_COMPONENT:
                return jsonify(self.run_handler(request.json).dump_handler())

            elif interaction_type == InteractionType.APPLICATION_COMMAND_AUTOCOMPLETE:
                return jsonify(self.run_autocomplete(request.json).dump())

    def set_route_async(self, route, app=None):
        """
        Add a route handler to a Quart app that handles incoming interaction
        data using asyncio.

        This function also sets up the aiohttp ClientSession that is used
        for sending followup messages, etc.

        Parameters
        ----------
        route
            The URL path to receive interactions on.
        app
            The Flask app to add the route to.
        """

        if app is None:
            app = self.app

        if aiohttp == None:
            raise ImportError(
                "The aiohttp module is required for async usage of this " "library"
            )

        @app.route(route, methods=["POST"])
        async def interactions():
            self.verify_signature(request)

            interaction_type = request.json.get("type")
            if interaction_type == InteractionType.PING:
                return jsonify({"type": ResponseType.PONG})

            elif interaction_type == InteractionType.APPLICATION_COMMAND:
                result = self.run_command(request.json)
            elif interaction_type == InteractionType.MESSAGE_COMPONENT:
                result = self.run_handler(request.json)
            elif interaction_type == InteractionType.APPLICATION_COMMAND_AUTOCOMPLETE:
                result = self.run_autocomplete(request.json)

            if inspect.isawaitable(result):
                result = await result

            return jsonify(result.dump())

        # Set up the aiohttp ClientSession

        async def create_session():
            app.discord_client_session = aiohttp.ClientSession(
                headers=self.auth_headers(app), raise_for_status=True
            )

        async def close_session():
            await app.discord_client_session.close()

        if hasattr(app, "before_serving"):
            # Quart apps
            app.before_serving(create_session)
            app.after_serving(close_session)
        else:
            # Flask apps
            app.before_first_request(create_session)
            atexit.register(
                lambda: asyncio.get_event_loop().run_until_complete(close_session())
            )



import warnings
from flask_discord_interactions.command import (
    Command,
    SlashCommandSubgroup,
    SlashCommandGroup,
)

from flask_discord_interactions.context import Context, AsyncContext

from flask_discord_interactions.models import (
    ApplicationCommandType,
    CommandOptionType,
    ChannelType,
    Permission,
    Member,
    User,
    Role,
    Channel,
    Message,
    ResponseType,
    Component,
    ActionRow,
    Button,
    ButtonStyles,
    ComponentType,
    SelectMenu,
    SelectMenuOption,
    Autocomplete,
    AutocompleteResult,
    Option,
)

from flask_discord_interactions.discord import (
    InteractionType,
    DiscordInteractions,
    DiscordInteractionsBlueprint,
)

import flask_discord_interactions.models.embed as embed
from flask_discord_interactions.models import Embed

from flask_discord_interactions.client import Client


class Response(Message):
    def __init__(self, *args, **kwargs):
        warnings.warn(
            "Deprecated! As of v1.1.0, Response has been renamed to Message, "
            "as it can now represent the argument to a Message Command.",
            DeprecationWarning,
            stacklevel=2,
        )
        return super().__init__(*args, **kwargs)


InteractionResponse = Response


class SlashCommand(Command):
    def __init__(self, *args, **kwargs):
        warnings.warn(
            "Deprecated! As of v1.1.0, SlashCommand has been renamed to Command, "
            'as it can represent ChatInput ("slash") commands, '
            "user commands, and message commands.",
            DeprecationWarning,
            stacklevel=2,
        )
        return super().__init__(*args, **kwargs)


class InteractionContext(Context):
    def __init__(self, *args, **kwargs):
        warnings.warn(
            "Deprecated! As of v0.1.5, "
            "InteractionContext has been renamed to Context.",
            DeprecationWarning,
            stacklevel=2,
        )
        return super().__init__(*args, **kwargs)


__all__ = [
    "embed",
    "Command",
    "SlashCommand",
    "SlashCommandSubgroup",
    "SlashCommandGroup",
    "Context",
    "AsyncContext",
    "CommandOptionType",
    "ApplicationCommandType",
    "ChannelType",
    "Member",
    "User",
    "Role",
    "Channel",
    "InteractionType",
    "DiscordInteractions",
    "DiscordInteractionsBlueprint",
    "Message",
    "ResponseType",
    "Embed",
    "Component",
    "ComponentType",
    "ActionRow",
    "Button",
    "ButtonStyles",
    "SelectMenu",
    "SelectMenuOption",
    "Client",
    "Permission",
    "Autocomplete",
    "AutocompleteResult",
    "Response",
    "InteractionResponse",
    "InteractionContext",
    "Option",
]
